import winreg
import os, json
from typing import List, Optional, Any
from dataclasses import dataclass
from zetsubou.project.model.platform_enums import EArch

from zetsubou.project.model.toolchain import ECompilerFamily, ToolchainDefinition
from zetsubou.system.windows import vswhere
from zetsubou.utils import logger
from zetsubou.utils.common import Version, get_env_path, get_files, get_subdirs
from zetsubou.utils.subprocess import call_process


def is_current_host(host: str, host_arch: EArch) -> bool:
    # x64
    if host.endswith('64'):
        if host_arch == EArch.x64:
            return True
    # x86 (32-bit)
    else:
        if host_arch != EArch.x64:
            return True
    return False


@dataclass
class Toolset:
    full_path: str
    identifier: str

    def __lt__(self, other):
        return self.identifier > other.identifier


def find_toolsets(msvc_path : str) -> List[Toolset]:
    toolsets = []
    search_subpath = 'Msbuild\\Microsoft\\VC'
    search_path = os.path.join(msvc_path, search_subpath)
    if os.path.exists(search_path):
        for root, _, files in os.walk(search_path):
            if 'Toolset.props' in files:
                identifier = os.path.basename(root)
                if identifier.startswith('v'):
                    toolsets.append(Toolset(full_path=root, identifier=identifier))

    return toolsets


def unique_toolsets(toolsets:List[Toolset]):
    result = set()
    for entry in toolsets:
        result.add(entry.identifier)

    return list(result)


def get_toolset_ver(toolset:Toolset):
    if toolset.identifier.startswith('v'):
        return int(toolset.identifier[1:])
    return 0


# Value defined in _MSVC_VER
def get_cl_exe_compiler_version(cl_exe_path:str):
    out, err = call_process([cl_exe_path], capture=True, realtime=False)
    if err is not None:
        return None

    # Strip prefix
    version = ''
    for i in range(len(out)):
        if out[i:i+1].isdigit():
            version = out[i:]
            break

    # Strip suffix
    version = version.split()[0]

    return version


# Filters out legacy versions of toolsets, leaving only the latest one
def filter_legacy_toolsets(msvc_path : str, toolsets: List[Toolset]) -> List[Toolset]:
    msvc_path = os.path.join(msvc_path, 'Msbuild\\Microsoft\\VC\\')
    selected_version = ''
    result = []

    # Filter out legacy toolsets
    for tool in toolsets:
        if tool.full_path.startswith(msvc_path):
            tmp_path = tool.full_path[len(msvc_path):]
            first_slash_idx = tmp_path.find('\\')
            tmp_path = tmp_path[:first_slash_idx]
            if tmp_path > selected_version:
                selected_version = tmp_path
    for tool in toolsets:
        if selected_version in tool.full_path:
            result.append(tool)

    return result


def get_registry_key(storage : str, key : str, entry : str) -> Optional[Any]:
    try:
        hkey = winreg.OpenKey(storage, key)
    except:
        return None
    else:
        try:
            value, _ = winreg.QueryValueEx(hkey, entry)
            return value
        except:
            return None
        finally:
            winreg.CloseKey(hkey)


@dataclass
class WindowsSDKPaths:
    version : str
    base_path : str
    include : str
    lib : str
    bin : str


def discover_win10_sdk():
    storages = [winreg.HKEY_LOCAL_MACHINE, winreg.HKEY_CURRENT_USER]
    key_paths = ['SOFTWARE\\Wow6432Node', 'SOFTWARE']

    def is_valid_sdk_ver(sdk_ver:str) -> bool:
        return sdk_ver.startswith('10.')

    for storage in storages:
        for key in key_paths:

            key_path = f'{key}\\Microsoft\\Microsoft SDKs\\Windows\\v10.0'
            install_dir = get_registry_key(storage, key_path, 'InstallationFolder')

            if install_dir and os.path.isdir(install_dir):
                include_dir = os.path.join(install_dir, 'include')

                sdk_versions = filter(is_valid_sdk_ver, os.listdir(include_dir))
                latest_sdk = max(sdk_versions)

                logger.Verbose(f"Possible Win10SDK {install_dir}\nlatest: {latest_sdk}\nall:{ ', '.join(sdk_versions) }")

                if os.path.isdir(os.path.join(include_dir, latest_sdk)) and latest_sdk.startswith('10.'):
                    windows_h = os.path.join(include_dir, latest_sdk, 'um', 'Windows.h')
                    if os.path.isfile(windows_h):
                        sdk = WindowsSDKPaths(
                            version=latest_sdk,
                            base_path=install_dir,
                            include=os.path.join(install_dir, 'Include', latest_sdk),
                            lib=os.path.join(install_dir, 'Lib', latest_sdk),
                            bin=os.path.join(install_dir, 'bin', latest_sdk),
                        )
                        return sdk
    return None


def discover_msvc(host_arch: EArch, vs_instances) -> List[ToolchainDefinition]:
    toolchains : List[ToolchainDefinition] = []

    default_vc_path = 'VC\\Tools\\MSVC'
    default_bin_path = 'bin'
    default_lib_path = 'lib'
    default_include_paths = ['include', 'atlmfc\\include']
    default_compiler_msvc = 'cl.exe'
    default_linker_msvc = 'link.exe'
    default_librarian_msvc = 'lib.exe'

    additional_include_dirs = [
        'VC\\Auxiliary\\VS\\include',
        'VC\\Auxiliary\\VS\\UnitTest\\include'
    ]

    default_extra_files = [
        'c1.dll',
        'c2.dll',
        'c1xx.dll',
        '1033\\clui.dll',
        'msobj140.dll',
        'mspdb140.dll',
        'mspdbsrv.exe',
        'mspdbcore.dll',
        'mspft140.dll',
        'msvcp140.dll',
        'vcruntime140.dll',
        'vccorlib140.dll'
    ]

    default_link_libraries = [
        'kernel32'
        'user32'
        'gdi32'
        'winspool'
        'comdlg32'
        'advapi32'
        'shell32'
        'ole32'
        'oleaut32'
        'uuid'
        'odbc32'
        'odbccp32'
        'delayimp'
    ]

    for vs_itr in vs_instances:
        installation_path = vs_itr['installationPath']
        vs_ide_version = vs_itr['installationVersion']
        vc_tools = os.path.join(installation_path, default_vc_path)
        vs_version = vs_itr['catalog']['productLineVersion']

        available_toolsets = unique_toolsets(find_toolsets(installation_path))

        if os.path.exists(vc_tools):

            vc_versions = get_subdirs(vc_tools)
            vc_ver = max(vc_versions)

            # for vc_ver in vc_versions:
            vc_ver_basepath = os.path.join(vc_tools, vc_ver)
            vc_ver_binpath = os.path.join(vc_ver_basepath, default_bin_path)
            vc_ver_libpath = os.path.join(vc_ver_basepath, default_lib_path)

            for host in get_subdirs(vc_ver_binpath):
                if is_current_host(host, host_arch):
                    for arch in get_subdirs(os.path.join(vc_ver_binpath, host)):

                        msvc_name = vc_ver.replace('.','_')
                        arch_bin_fullpath = os.path.join(vc_ver_binpath, host, arch)
                        arch_lib_fullpath = os.path.join(vc_ver_libpath, arch)

                        msvc_ver = get_cl_exe_compiler_version(os.path.join(arch_bin_fullpath, default_compiler_msvc))
                        if msvc_ver is None:
                            continue

                        toolchain_name = f'{ECompilerFamily.MSVC.name}_{arch}_vs{vs_version}_v{msvc_name}'

                        includes = []
                        includes += [ os.path.join(vc_ver_basepath, inc) for inc in default_include_paths ]
                        includes += [ os.path.join(installation_path, inc) for inc in additional_include_dirs ]

                        # TODO: maybe this should be separate for compiler, librarian and linker?
                        path_envs = []
                        path_envs += [ arch_bin_fullpath ]

                        machine_linker_options = f'/MACHINE:{arch} '

                        compiler = ToolchainDefinition.create(
                            toolchain_name,
                            os.path.join(arch_bin_fullpath, default_compiler_msvc),
                            os.path.join(arch_bin_fullpath, default_librarian_msvc),
                            os.path.join(arch_bin_fullpath, default_linker_msvc),

                            arch=arch,

                            version=Version(msvc_ver),
                            ide_version=Version(vs_ide_version),
                            vs_install_path=installation_path,

                            Toolset=available_toolsets,

                            PathEnv=path_envs,

                            LinkerOptions=machine_linker_options,

                            LinkerPaths=[
                                arch_lib_fullpath,
                                arch_bin_fullpath,
                            ],

                            ExtraFiles = default_extra_files,

                            RequiredFiles=[],
                            IncludeDirectories=includes,
                            Defines=[]
                        )
                        toolchains.append(compiler)

    return toolchains


def get_clang_version(clang_exe_path: str) -> Optional[str]:
    out, err = call_process([clang_exe_path, '--version'], capture=True, realtime=False)
    if err is not None:
        return None

    version_prefix = 'clang version '
    version = out[len(version_prefix):]
    version = version.split()[0]

    return version


def discover_clang(host_arch:EArch, vs_instances) -> List[ToolchainDefinition]:
    # We need to find visual studio installations anyway
    # But how to handle standalone LLVM? Ignore it? Or match with what version?
    # All of them? The latest? Per toolset?

    # There is a thing I cannot find documentation about
    # Seems the Clang toolset exists in two versions per visual studio installation
    #   egx. 2022 has v170 and v160 folders with Clang
    #    and 2019 has v160 and v150
    # We don't want th
    # toolsets = filter_legacy_toolsets(installation_path, toolsets)

    # Find VS installations
    # Find Clang from visual studios

    # Find standalone LLVM installation
    # Match with vs toolsets

    # clang.exe will work as an compiler, librarian and linker
    # but we need link.exe to be present in path when calling the compiler
    # how on earth would we do that in fastbuild..?
    # Compiler in fastbuild has environment option!

    # TODO: We need to pass Environment from Compiler to DLLs, Libraries and Execs!
    # This will set correct PATH with linker.exe
    toolchains : List[ToolchainDefinition] = []

    default_msvc_path = 'VC\\Tools\\MSVC'
    default_llvm_path = 'VC\\Tools\\Llvm'
    default_bin_path = 'bin'
    default_lib_path = 'lib'
    default_llvm_include_path = 'include'
    default_msvc_include_paths = ['include', 'atlmfc\\include']
    default_compiler_clang = 'clang.exe'
    default_librarian_clang = 'llvm-ar.exe'
    default_linker_clang = 'lld-link.exe'
    default_compiler_msvc = 'cl.exe'
    default_linker_msvc = 'link.exe'

    additional_include_dirs = [
        'VC\\Auxiliary\\VS\\include',
        'VC\\Auxiliary\\VS\\UnitTest\\include'
    ]

    default_extra_files = [
        'c1.dll',
        'c2.dll',
        'c1xx.dll',
        '1033\\clui.dll',
        'msobj140.dll',
        'mspdb140.dll',
        'mspdbsrv.exe',
        'mspdbcore.dll',
        'mspft140.dll',
        'msvcp140.dll',
        'vcruntime140.dll',
        'vccorlib140.dll'
    ]

    standalone_clangs = []

    def try_create_clang_toolchain(clang_exe_path, msvc_ver_basepath, msvc_host, msvc_arch, vs_version, toolset:Toolset) -> Optional[ToolchainDefinition]:
        clang_ver = get_clang_version(clang_exe_path)
        if clang_ver is None:
            return None

        vc_ver_binpath = os.path.join(msvc_ver_basepath, default_bin_path)
        vc_ver_libpath = os.path.join(msvc_ver_basepath, default_lib_path)
        msvc_arch_bin_fullpath = os.path.join(vc_ver_binpath, msvc_host, msvc_arch)
        msvc_arch_lib_fullpath = os.path.join(vc_ver_libpath, msvc_arch)
        msvc_linker_exe = os.path.join(msvc_arch_bin_fullpath, default_linker_msvc)

        msvc_ver = get_cl_exe_compiler_version(os.path.join(msvc_arch_bin_fullpath, default_compiler_msvc))
        if msvc_ver is None:
            return None

        clang_ver_name = clang_ver.replace('.','_')

        clang_bin_dir = os.path.dirname(clang_exe_path)
        clang_base_dir = os.path.dirname(clang_bin_dir)
        clang_lib_dir = os.path.join(clang_base_dir, default_lib_path)
        clang_include_dir = os.path.join(clang_base_dir, default_llvm_include_path)
        clang_librarian_exe = os.path.join(clang_bin_dir, default_librarian_clang)
        clang_linker_clang = os.path.join(clang_bin_dir, default_linker_clang)

        compiler_name = f"{ECompilerFamily.CLANG.name}_{msvc_arch}_vs{vs_version}_v{clang_ver_name}"

        includes = [ clang_include_dir ]
        includes += [ os.path.join(msvc_ver_basepath, inc) for inc in default_msvc_include_paths ]
        includes += [ os.path.join(installation_path, inc) for inc in additional_include_dirs ]

        # TODO: maybe this should be separate for compiler, librarian and linker?
        path_envs = []
        path_envs += [ clang_bin_dir, msvc_arch_bin_fullpath ]

        machine_linker_options = f'/MACHINE:{msvc_arch} '

        sem_msvc_ver = Version(msvc_ver)
        msvc_compat_options = f"-fms-compatibility-version={sem_msvc_ver.major}{sem_msvc_ver.minor} "

        return ToolchainDefinition.create(
            compiler_name,
            clang_exe_path,
            clang_librarian_exe,
            msvc_linker_exe,

            CompilerOptions = msvc_compat_options,
            LinkerOptions = machine_linker_options,

            arch = msvc_arch,
            version = Version(clang_ver),
            ide_version = Version(vs_ide_version),
            vs_install_path = installation_path,

            Toolset = [ toolset.identifier ],
            PathEnv = path_envs,
            ExtraFiles = default_extra_files,

            LinkerPaths = [
                clang_lib_dir,
                clang_bin_dir,
                msvc_arch_lib_fullpath,
                msvc_arch_bin_fullpath,
            ],

            RequiredFiles=[],
            IncludeDirectories=includes,
            Defines=[]
        )

    env_path = get_env_path()
    for path in env_path:
        for file in get_files(path): 
            if file.find(default_compiler_clang) != -1:
                clang_path = os.path.join(path, file)
                standalone_clangs.append(clang_path)

    for vs_itr in vs_instances:
        installation_path = vs_itr['installationPath']
        vs_ide_version = vs_itr['installationVersion']
        msvc_tools = os.path.join(installation_path, default_msvc_path)
        llvm_tools = os.path.join(installation_path, default_llvm_path)
        vs_version = vs_itr['catalog']['productLineVersion']

        toolsets = find_toolsets(installation_path)
        latest_toolset = max(toolsets, key=get_toolset_ver)

        logger.Verbose(f"For VS {installation_path} picked {latest_toolset} from found toolsets:\n{[t.identifier for t in toolsets]}")

        msvc_versions = get_subdirs(msvc_tools)
        latest_msvc = max(msvc_versions)

        # for vc_ver in vc_versions:
        msvc_ver_basepath = os.path.join(msvc_tools, latest_msvc)
        vc_ver_binpath = os.path.join(msvc_ver_basepath, default_bin_path)

        if os.path.exists(msvc_tools):

            for msvc_host in get_subdirs(vc_ver_binpath):
                if is_current_host(msvc_host, host_arch):
                    for msvc_arch in get_subdirs(os.path.join(vc_ver_binpath, msvc_host)):

                        # Standalone Clang
                        for clang_entry in standalone_clangs:
                            compiler = try_create_clang_toolchain(clang_entry, msvc_ver_basepath, msvc_host, msvc_arch, vs_version, latest_toolset)
                            if compiler is None:
                                continue

                            toolchains.append(compiler)

                        # Bundled Clang
                        if os.path.exists(llvm_tools):
                            for host in get_subdirs(llvm_tools):
                                if is_current_host(host, host_arch):
                                    if host == default_bin_path:
                                        host = ''

                                    clang_dir = os.path.join(llvm_tools, host, default_bin_path)
                                    clang_exe = os.path.join(clang_dir, default_compiler_clang)

                                    compiler = try_create_clang_toolchain(clang_exe, msvc_ver_basepath, msvc_host, msvc_arch, vs_version, latest_toolset)
                                    if compiler is None:
                                        continue

                                    toolchains.append(compiler)

    return toolchains


def discover_toolchain_list(host_arch: EArch) -> List[ToolchainDefinition] :
    vswhere_path = None

    try:
        vswhere_path = vswhere.find_vswhere()
    except Exception:
        pass

    if vswhere_path is None:
        logger.Warning("Unable to find vswhere executable, VS toolchains not discovered")
        return []

    vs_instances = json.loads(vswhere.call_vswhere(vswhere_path))
    if vs_instances is None:
        logger.Warning("vswhere found no VS installations")
        return []

    toolchains : List[ToolchainDefinition] = []
    toolchains += discover_msvc(host_arch, vs_instances)
    toolchains += discover_clang(host_arch, vs_instances)

    win10_sdk = discover_win10_sdk()

    if win10_sdk is not None:
        logger.Verbose(f"Found Win10SDK in {win10_sdk.base_path}")
        for toolchain in toolchains:
            # This should work for clang/clangCL too
            #if toolchain.CompilerFamily == ECompilerFamily.MSVC:
            toolchain.LinkerPaths.append(f'{win10_sdk.lib}\\um\\{toolchain.arch.lower()}')
            toolchain.LinkerPaths.append(f'{win10_sdk.lib}\\ucrt\\{toolchain.arch.lower()}')
            toolchain.IncludeDirectories.append(f'{win10_sdk.include}\\shared')
            toolchain.IncludeDirectories.append(f'{win10_sdk.include}\\um')
            toolchain.IncludeDirectories.append(f'{win10_sdk.include}\\ucrt')

    return toolchains



import json, dataclasses
from enum import Enum

class EnhancedJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        if isinstance(o, Enum):
            return o.name
        return super().default(o)

def to_json(obj):
    return json.dumps(obj, cls=EnhancedJSONEncoder, indent=4, sort_keys=False)



import ctypes
import sys
from typing import List
from zetsubou.project.model.platform_enums import ESystem, EArch
from zetsubou.project.model.platform import Platform


def get_host_system() -> ESystem:
    if sys.platform == 'win32':
        return ESystem.Windows
    if sys.platform in [ 'linux', 'linux2' ]:
        return ESystem.Linux
    raise EnvironmentError(f'Sorry, \"{sys.platform}\" is not currently supported!')


def get_host_arch() -> EArch:
    # More architectures might be needed here
    if is_64_bit():
        return EArch.x64
    return EArch.x86


def filter_platforms_by_host(host_system: ESystem, host_arch: EArch, platforms: List[Platform]) -> List[Platform]:
    results = []
    for plat in platforms:
        if plat.host_arch == host_arch and plat.host_system == host_system:
            results.append(plat)
    return results


def is_64_bit():
    return ctypes.sizeof(ctypes.c_void_p) == 8



from copy import copy
from zetsubou.commands.base_command import CommandContext
from typing import Set, Tuple, List
from zetsubou.project.model.configuration import Configuration
from zetsubou.project.model.configuration_enums import EBaseConfiguration
from zetsubou.project.model.platform_enums import EArch, ESystem
from zetsubou.project.model.platform import Platform
from zetsubou.project.model.runtime_library import ERuntimeLibrary
from zetsubou.project.model.toolchain_profile import Windows_ToolchainProfile
from zetsubou.project.model.toolchain import ECompilerFamily, ECppStandard, Toolchain
from zetsubou.utils.common import Version


def platform_to_conan(context: CommandContext, plat : Platform) -> List[str]:
    return [
        '-s', f'os={plat.host_system.name}'
    ]


def msvc_version_to_conan(msvc_ver:Version) -> Tuple[str, str]:
    version = str(msvc_ver.major)
    version += str(msvc_ver.minor)[0]
    return (version, str(msvc_ver.minor)[1:])


def compiler_family_to_conan(family : ECompilerFamily):
    return {
        ECompilerFamily.MSVC.name  : 'msvc',
        ECompilerFamily.CLANG.name : 'clang',
        ECompilerFamily.GCC.name   : 'gcc'
    }.get(family.name)


def toolchain_to_conan(context: CommandContext, toolchain: Toolchain) -> List[str]:
    if toolchain is not None:

        arch = {
            EArch.x86.name : 'x86',
            EArch.x64.name : 'x86_64'
        }.get(toolchain.definition.arch)

        compiler = compiler_family_to_conan(toolchain.definition.CompilerFamily)

        result = [
            '-s', f'arch={arch}',
            '-s', f'compiler={compiler}',
        ]

        # TODO FIXME Conan should know about these, but there is no way to pass them via CLI anymore in 2.0 :(
        # for path in toolchain.definition.PathEnv:
        #     result.extend([
        #         '-e', f'PATH=[{path}]'
        #     ])

        result.extend(cppstd_to_conan(toolchain.profile.cppstd))

        if context.host_system == ESystem.Windows and toolchain.toolset is not None:
            if toolchain.definition.CompilerFamily == ECompilerFamily.CLANG:
                result.extend([
                    '-s', f'compiler.runtime_version={toolchain.toolset}'
                ])

        if toolchain.definition.CompilerFamily == ECompilerFamily.MSVC:
            version = msvc_version_to_conan(toolchain.definition.version)
            result.extend([
                '-s', f'compiler.version={version[0]}',
                '-s', f'compiler.update={version[1]}'
            ])
        else:
            # Remove path, conan is interested only in Major.Minor at best
            ver_no_path = copy(toolchain.definition.version)
            ver_no_path.path = 0

            result.extend([
                '-s', f'compiler.version={ver_no_path}'
            ])

        return result

    return []


def config_base_to_conan(context: CommandContext, config: Configuration) -> List[str]:
    if config is not None:

        build_type = {
            EBaseConfiguration.DEBUG : 'Debug',
            EBaseConfiguration.RELEASE : 'Release',
            EBaseConfiguration.RELEASE_WITH_DEBUG_INFO : 'RelWithDebInfo'
        }.get(config.base_configuration)

        result = [
            '-s', f'build_type={build_type}',
        ]

        return result

    return []


def cppstd_to_conan(cppstd:ECppStandard):
    cpp = {
        ECppStandard.cpp11  : '11',
        ECppStandard.cpp14  : '14',
        ECppStandard.cpp17  : '17',
        ECppStandard.cpp20  : '20',
        ECppStandard.cpp23  : '23',
        ECppStandard.latest : '23',
    }.get(cppstd)
    return [
        '-s', f"compiler.cppstd={cpp}"
    ]


def runtime_to_conan(config:Configuration, toolchain:Toolchain):
    if not isinstance(toolchain.profile, Windows_ToolchainProfile):
        return []

    runtime = {
        ERuntimeLibrary.dynamic : 'dynamic',
        ERuntimeLibrary.static : 'static'
    }.get(toolchain.profile.runtime)

    runtime_type = 'Debug' if config.base_configuration == EBaseConfiguration.DEBUG else 'Release'

    return [
        '-s', f'compiler.runtime={runtime}',
        '-s', f'compiler.runtime_type={runtime_type}'
    ]


def toolsets_from_settings(conan_settings : dict, compiler_family : ECompilerFamily) -> Set[str]:
    conan_compiler = compiler_family_to_conan(compiler_family)
    if compiler_family == ECompilerFamily.CLANG:
        try:
            return set(conan_settings['compiler'][conan_compiler]['runtime_version']) if conan_settings is not None else None
        except KeyError:
            return None
    else:
        return None



from dataclasses import dataclass
from typing import Optional
from fs.base import FS
from zetsubou.project.model.platform_enums import ESystem, EArch
from zetsubou.project.model.virtual_environment import VirtualEnvironment


@dataclass
class BaseContext:
    host_system: ESystem = 0
    host_arch: EArch = 0
    fs_root: str = ''
    fs_venv: Optional[VirtualEnvironment] = None
    project_fs: FS = ''
    project_file: str = ''



# Simple yaml writer with no dependencies
# Can be used in Conan generator
from typing import List
from enum import Enum


def is_enum(obj):
    if obj is None:
        return False
    return isinstance(obj, Enum)


def is_obj(obj):
    if obj is None:
        return False
    return hasattr(obj, '__dict__')


def is_list(obj):
    if obj is None:
        return False
    return isinstance(obj, list)


class _yaml_context:
    buffer: List[str] = []
    cur_indent: int = 0
    DEF_INDENT = 2
    prefix_once = None

    def push_obj(self, obj):
        self.cur_indent += self.DEF_INDENT

    def pop_obj(self, obj):
        self.cur_indent -= self.DEF_INDENT

    def assign(self, key:str, val=None):
        indent = self.cur_indent
        prefix = ''

        if self.prefix_once is not None:
            prefix = self.prefix_once
            self.prefix_once = None
            indent -= len(prefix)

        if val is None:
            self.buffer.append(f'{" " * indent}{prefix}{key}:')
        else:
            self.buffer.append(f'{" " * indent}{prefix}{key}: {str(val)}')


def _handle_key_val(ctx:_yaml_context, key:str, val):
    if val is None:
        return
    if is_enum(val):
        ctx.assign(key, val.name)
    elif is_list(val):
        if len(val) == 0:
            return
        ctx.assign(key)
        ctx.push_obj(val)
        for sub_obj in val:
            ctx.prefix_once = '- '
            _handle_obj(ctx, sub_obj)
        ctx.pop_obj(val)
    elif is_obj(val):
        ctx.assign(key)
        ctx.push_obj(val)
        _handle_obj(ctx, val)
        ctx.pop_obj(val)
    else:
        ctx.assign(key, val)


def _handle_obj(ctx:_yaml_context, obj):
    for name, value in vars(obj).items():
        _handle_key_val(ctx, name, value)


def to_yaml(obj) -> str:
    ctx = _yaml_context()
    _handle_obj(ctx, obj)
    return '\n'.join(ctx.buffer)



# pyright: reportMissingImports=false
# pylint: disable=import-error, bare-except
import os, sys
import yaml
from typing import Optional

from conan import ConanFile
from conan.errors import ConanException
from conan.tools.files import copy
from conan.tools.env import VirtualBuildEnv
from conans.client.loader_txt import ConanFileTextLoader


toolchain_available:bool = False
sys.path.insert(0, os.path.dirname(__file__))

try:
    from zetsubou.conan import from_conan
    from zetsubou.conan import deployer
    toolchain_available = True
except Exception as ex:
    print(f"Failed to import 'from_conan' {ex}")
try:
    import from_conan
    import deployer
    toolchain_available = True
except Exception as ex:
    print(f"Failed to import 'from_conan' {ex}")


# ZetsubouToolchain
# Provides methods to extract dependency information from project.yml
# Can generate the project files and build it
class _toolchain(object):
    project_file: str
    conanfile: ConanFile
    platform_file = None

    _project_template = None
    _conan_dependencies = None

    def _load_project_template(self, project_yml:str):
        if not os.path.exists(project_yml):
            return None
        if self._project_template is None:
            with open(project_yml, mode='r') as in_proj:
                self._project_template = yaml.load(in_proj, yaml.SafeLoader)
        return self._project_template

    def _load_conan_ini(self, conan_ini:Optional[str]):
        if conan_ini is None:
            return None

        if not os.path.exists(conan_ini) or not os.path.isfile(conan_ini):
            return None

        with open(conan_ini, mode='r') as in_ini:
            conan_ini_content = in_ini.read()
            return ConanFileTextLoader(conan_ini_content)

    def _load_conan_dependencies(self, project_yml:str):
        if self._conan_dependencies is None:
            proj = self._load_project_template(project_yml)
            if proj is None:
                return None

            conan_data = proj.get('conan', None)
            if conan_data is None:
                return None

            tools_req_file = conan_data.get('build_tools', None)
            req_file = conan_data.get('dependencies', None)

            result = {}

            tools_ini = self._load_conan_ini(tools_req_file)
            if tools_ini is not None:
                result['tool_requirements'] = tools_ini.tool_requirements

            main_ini = self._load_conan_ini(req_file)
            if main_ini is not None:
                result['requirements'] = main_ini.requirements

            self._conan_dependencies = result
        return self._conan_dependencies

    def fill_requirements(self, conanfile:ConanFile, project_yml:str):
        for entry in self._load_conan_dependencies(project_yml).get('requirements', []):
            if not any(filter(lambda r : r.ref.matches(entry, is_consumer=False), conanfile.requires.values())):
                conanfile.requires(entry)

    def fill_tool_requirements(self, conanfile:ConanFile, project_yml:str):
        for entry in self._load_conan_dependencies(project_yml).get('tool_requirements', []):
            if not any(filter(lambda r : r.ref.matches(entry, is_consumer=False), conanfile.requires.values())):
                conanfile.requires.tool_require(entry)

    def profile_filename(self, dir:str):
        return os.path.join(dir, 'conan_profile.yml')

    def init(self, conanfile:ConanFile, project_file:str="project.yml"):
        self.conanfile = conanfile
        self.project_file = project_file

    def configure(self):
        if not self.conanfile:
            raise ValueError('Zetsubou not initialized! call self.zetsubou.init(self) before using!')

        if toolchain_available:
            profile_file = self.profile_filename(self.conanfile.folders.generators_folder)
            if not from_conan.export_profile(self.conanfile, profile_file):
                raise ConanException('Unable to generate profile file!')

            # Previously, we were calling zetsubou config, but we shouldn't call install from within conan
            # Instead we generate virtual env here
            venv = VirtualBuildEnv(self.conanfile)
            venv.generate()

            # And generate .yml files for conan dependencies here
            deployer.generate(self.conanfile)

            # And consume it by just generating the project files
            self.conanfile.run(f'zetsubou gen {self.project_file} --profile={profile_file} --nologo --ide')
        else:
            raise ConanException('Unable to start Zetsubou, failed imports!')

    def build(self):
        if not self.conanfile:
            raise ValueError('Zetsubou not initialized! call self.zetsubou.init(self) before using!')

        if not toolchain_available:
            raise ConanException('Unable to start Zetsubou, failed imports!')

        profile_file = self.profile_filename(self.conanfile.folders.generators_folder)
        if not os.path.exists(profile_file) or not os.path.isfile(profile_file):
            raise ConanException(f"Unable to find '{profile_file}' profile file! Did you forget to run install first?")

        self.conanfile.run(f'zetsubou build {self.project_file} --profile={profile_file} --nologo --ide')

# Conanfiles using Zetsubout should use this class as a base
# It's purpose is to read project.yml file and inject dependencies 
# Into conanfile, without a need to setup it in two places
class ZetsubouBase(object):
    _zetsubou: _toolchain = _toolchain()
    _project_yml: str = ''

    @property
    def zetsubou(self):
        if self._zetsubou is None:
            print("Zetsubou not initialized!")
        return self._zetsubou

    # This is unfortunately a hack
    # We store dependency data outside conanfile.py and need to read and setup them here
    # What makes life harder is a fact that folder paths are being setuped after dependency graph
    # So we need to guess where hardcoded project.yml is based on recipe path
    # TODO Might not work for Conan 2.0!
    @property
    def default_project(self):
        if self._project_yml == '':
            from_recipe_folder = os.path.join(self.recipe_folder, 'project.yml')

            if os.path.exists(from_recipe_folder) and os.path.isfile(from_recipe_folder):
                self._project_yml = from_recipe_folder
            elif self.recipe_folder.endswith('e'):
                from_export = os.path.normpath(os.path.join(self.recipe_folder, '..', 'es', 'project.yml'))
                if os.path.exists(from_export) and os.path.isfile(from_export):
                    self._project_yml = from_export

        return self._project_yml

    def requirements(self):
        self.zetsubou.fill_requirements(self, self.default_project)

    def build_requirements(self):
        self.zetsubou.fill_tool_requirements(self, self.default_project)


class ZetsubouGeneratorPackage(ConanFile):
    name = "zetsubougen"
    version = "0.7.0"
    license = "MIT"
    description = 'Zetsubou yml files generator'
    url = "https://github.com/BentouDev/Zetsubou"
    package_type = "python-require"

    # Must be placed at the root of the source
    # Conan no longer allows referencing parent directory in relative paths
    def export(self):
        exports = [
            "conanfile.py",
            "conan/from_conan.py",
            "conan/deployer.py",
            "project/model/toolchain_profile.py",
            "project/model/platform_enums.py",
            "project/model/configuration_enums.py",
            "project/model/toolchain_enums.py",
            "project/model/runtime_library.py",
            "project/model/kind.py",
            "utils/yaml_simple_writer.py",
        ]

        for e in exports:
            copy(self, e, self.recipe_folder, self.export_folder, keep_path=False)



from typing import List, Optional
from dataclasses import dataclass, field

from zetsubou.project.model.filter import TargetFilter
from zetsubou.project.model.kind import ETargetKind
from bentoudev.dataclass.base import inline_loader, track_source, loaded_from_file


@dataclass
class Dependencies:
    interface: Optional[List['zetsubou.project.model.target.TargetReference']] = field(default_factory=list) # noqa: F821
    public: Optional[List['zetsubou.project.model.target.TargetReference']] = field(default_factory=list)    # noqa: F821
    private: Optional[List['zetsubou.project.model.target.TargetReference']] = field(default_factory=list)   # noqa: F821


@dataclass
class Source:
    paths: List[str] = field(default_factory=list)
    patterns: Optional[List[str]] = field(default_factory=list)


@dataclass
class PropertyList:
    interface: Optional[List[str]] = field(default_factory=list)
    public: Optional[List[str]] = field(default_factory=list)
    private: Optional[List[str]] = field(default_factory=list)


@dataclass
class PathList(PropertyList):
    pass


@dataclass
class TargetData:
    dependencies: Optional[Dependencies] = field(default=None)
    source: Optional[Source] = field(default=None)
    source_exclude: Optional[Source] = field(default=None)
    includes: Optional[PathList] = field(default=None)
    system_includes: Optional[PathList] = field(default=None)
    defines: Optional[PropertyList] = field(default=None)
    compiler_flags: Optional[PropertyList] = field(default=None)
    linker_flags: Optional[PropertyList] = field(default=None)
    librarian_flags: Optional[PropertyList] = field(default=None)
    linker_paths: Optional[PathList] = field(default=None)
    link_libraries: Optional[PropertyList] = field(default=None)
    distribute_files: Optional[Source] = field(default=None)


@dataclass
class TargetFilterData(TargetData):
    filter: TargetFilter = field(default_factory=TargetFilter)


@dataclass
class TargetConfig:
    kind: ETargetKind
    compiler: Optional[str] = None


@dataclass
@loaded_from_file
@track_source
class Target(TargetData):
    target: str = ''
    config: TargetConfig = None
    filters: Optional[List[TargetFilterData]] = field(default_factory=list)


@inline_loader(source_type=str, field_name='name')
@track_source
@dataclass
class TargetReference:
    name: str
    target: Target = None

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if isinstance(other, TargetReference):
            return self.name == other.name
        return False

    def __ne__(self, other):
        # Not strictly necessary, but to avoid having both x==y and x!=y
        # True at the same time
        return not(self == other)


def find_target(targets: List[Target], name: str) -> Optional[Target]:
    for target in targets:
        if target.target == name:
            return target
    return None



from __future__ import absolute_import
import os
import time
from typing import Optional, List
from sys import argv as sys_argv

from fs.osfs import OSFS
from zetsubou.commands.base_command import CommandContext, Command
from zetsubou.commands.command_registry import get_all_commands

from zetsubou.logo import print_logo
from zetsubou.project.runtime.project_loader import ProjectError
from zetsubou.utils import logger
from zetsubou.utils.cmd_arguments import parse_arguments
from zetsubou.utils.common import fix_path
from zetsubou.utils.error import CompilationError
from zetsubou.utils.error_codes import EErrorCode, ReturnErrorcode
from zetsubou.system.tools import get_host_arch, get_host_system
from zetsubou._version import get_author_desc


def main(argv: List[str] = sys_argv[1:]) -> int:
    desc = f'FASTbuild project generator for the helpless\n{get_author_desc()}'
    progname = 'zetsubou'
    start_time = time.time()
    logger.Initialize()

    try:
        # populate command list
        command_registry = get_all_commands()

        # parse cmd arguments
        zet_args = parse_arguments(argv, progname, desc, command_registry)

        if not zet_args.nologo and not zet_args.ide and not zet_args.silent:
            print_logo(desc)

        if zet_args.ide:
            logger.SetIde(True)

        if zet_args.silent:
            logger.SetLogLevel(logger.ELogLevel.Silent)
        elif zet_args.verbose:
            logger.SetLogLevel(logger.ELogLevel.Verbose)

        has_project_context = hasattr(zet_args, 'project')

        fs_root = os.path.dirname(os.path.normpath(os.path.join(os.getcwd(), zet_args.project))) if has_project_context else os.getcwd()
        project_file = os.path.basename(zet_args.project) if has_project_context else None

        command_context = CommandContext(
            host_arch = get_host_arch(),
            host_system = get_host_system(),
            fs_root = fix_path(fs_root),
            command_args = zet_args,
            project_file = project_file,
            project_fs = OSFS(fs_root)
        )

        logger.Info(f"Current working directory - '{os.getcwd()}'")
        logger.Info(f"Project working directory - '{fs_root}'")

        cmd = Command.get_command_by_name(zet_args.command)
        cmd.Execute(command_context)

        end_time = time.time()

        if not zet_args.silent:
            print(f'\nFinished in {end_time - start_time:.2f} sec')

        return 0

    except ReturnErrorcode as errcode:
        logger.ReturnCode(errcode.error_code)
        return errcode.error_code.value

    except ProjectError as proj_error:
        logger.Error(proj_error)
        logger.ReturnCode(EErrorCode.UNKNOWN_ERROR)
        return EErrorCode.UNKNOWN_ERROR.value

    except CompilationError:
        logger.ReturnCode(EErrorCode.COMPILATION_FAILED)
        return EErrorCode.COMPILATION_FAILED.value

    except Exception as error:
        logger.Exception(error)
        return EErrorCode.UNKNOWN_ERROR.value



from zetsubou.project.model.toolchain import ECompilerFamily


def compiler_family_to_fastbuild(compiler : ECompilerFamily):
    compiler_family = {
         ECompilerFamily.MSVC : 'msvc',
         ECompilerFamily.CLANG : 'clang',
         ECompilerFamily.GCC : 'gcc',
         ECompilerFamily.CUSTOM : 'custom',
    }

    if compiler in compiler_family:
        return compiler_family.get(compiler)

    return 'auto'



from enum import Enum, auto

# For compatibility purposes, custom configurations must fallback to one of the default ones, to communicate with VS and Conan
class EBaseConfiguration(Enum):
    DEBUG = auto()
    RELEASE = auto()
    RELEASE_WITH_DEBUG_INFO = auto()



from enum import Enum


class EErrorCode(Enum):
    UNABLE_TO_LOAD_PROJECT = 100
    UNABLE_TO_FIND_VENV = 150
    UNABLE_TO_DISCOVER_TOOLCHAIN = 200
    UNABLE_TO_INSTALL_CONAN_BUILD_TOOLS = 220
    UNABLE_TO_INSTALL_CONAN_DEPENDENCIES = 222
    UNABLE_TO_LOAD_DEPENDENCIES = 280
    UNABLE_TO_VALIDATE_TARGETS = 290
    UNABLE_TO_CREATE_CONFIGURATIONS = 300
    UNABLE_TO_RESOLVE_CLI_TOOLS_FULLPATH = 350
    UNABLE_TO_RESOLVE_VARIANTS = 400
    UNABLE_TO_EMIT_PROJECT = 500
    UNABLE_TO_COMMIT_FS = 550
    UNABLE_TO_CREATE_VS_PROJECTS = 600
    UNABLE_TO_CREATE_VS_SOLUTION = 650
    UNABLE_TO_CLEAN = 700
    COMPILATION_FAILED = 800
    UNKNOWN_ERROR = 666


class ReturnErrorcode(Exception):
    error_code: EErrorCode = 0x0

    def __init__(self, code: EErrorCode):
        self.error_code = code



from argparse import ArgumentParser
import os
from fs.base import FS
from fs.copy import copy_fs
from zetsubou.commands.base_command import Command
from zetsubou.commands.command_context import CommandContext, Fastbuild
from zetsubou.commands.execute_stage import execute_stage
from zetsubou.commands.install import Install
from zetsubou.conan.dependencies import ConanDependencies
from zetsubou.fastbuild.fastbuild_emit import FastbuildEmitter
from zetsubou.project.model.target import Target
from zetsubou.project.model.virtual_environment import VirtualEnvironment
from zetsubou.project.runtime.emit import EmitContext, emit_project
from zetsubou.project.runtime.project_loader import load_dataclass_list
from zetsubou.project.runtime.resolve import ResolveContext, resolve_target_variants
from zetsubou.project.runtime.validation import ValidationContext, validate_targets, validate_cli_tools

from zetsubou.utils import logger
from zetsubou.utils.error_codes import EErrorCode
from zetsubou.utils.subprocess import call_process_venv
from zetsubou.utils.yaml_loader import YamlDataclassLoader as YAMLLoader
import bentoudev.dataclass.base as base

from fs import open_fs as fs_create

def commit_to_filesystem(project_fs : FS, mem_fs : FS, fs_root : str):
    copy_fs(mem_fs, project_fs)
    return True


def check_fastbuild_present(venv : VirtualEnvironment):
    out, err = call_process_venv(['fbuild', '-version'], venv, capture=True, realtime=False)
    if err is not None:
        logger.CriticalError('FASTbuild is missing, unable to proceed')
        return False

    logger.Success("FASTbuild present")
    logger.Verbose(out)
    return True


def call_fbuild_projects(fs_root : dir, bff_dir : str, bff_file : str, venv_obj : VirtualEnvironment):
    bff_dir = os.path.join(fs_root, bff_dir)
    bff_file = os.path.join(fs_root, bff_dir, bff_file)

    verbose = logger.IsVisible(logger.ELogLevel.Verbose)

    assert os.path.isfile(bff_file)
    assert not os.path.isfile(bff_dir)

    out, err = call_process_venv(['fbuild', '-config', bff_file], venv_obj, capture = not verbose, realtime = verbose, cwd = bff_dir)
    if err is not None:

        if not verbose:
            logger.Error(out)

        logger.CriticalError(f"FASTbuild returned error '{err}'")
        return False

    if verbose:
        print('')
    return True


def call_fbuild_solution(fs_root : dir, bff_dir : str, bff_file : str, project_name : str, venv_obj : VirtualEnvironment):
    bff_dir = os.path.join(fs_root, bff_dir)
    bff_file = os.path.join(fs_root, bff_dir, bff_file)

    verbose = logger.IsVisible(logger.ELogLevel.Verbose)
    if verbose:
        print('')

    assert os.path.isfile(bff_file)
    assert not os.path.isfile(bff_dir)

    out, err = call_process_venv(['fbuild', '-config', bff_file, project_name, '-ide'], venv_obj, capture=not verbose, realtime=verbose, cwd=bff_dir)
    if err is not None:

        if not verbose:
            logger.Error(out)

        logger.CriticalError(f"FASTbuild returned error '{err}'")
        return False

    if verbose:
        print('')
    return True


def load_dependencies(context: CommandContext):
    loader = YAMLLoader()
    local_types = base.get_types_from_modules([__name__, 'zetsubou.project.model.target'])
    dep_targets = []

    conan_deps = context.project_template.project.conan.dependencies
    if conan_deps is None:
        return

    conan_deps = context.to_out_path(conan_deps)

    def load_deps():
        path = 'build/conan_deps.yml'

        # Merge files, only parts from deployer exist
        if not context.project_fs.exists(path):
            Command.get_command_instance(Install).merge_generated_part_files(context, conan_deps)
            return True

        # Load merged yaml
        with context.project_fs.open(path, 'r', encoding='utf-8') as obj_file:
            obj_templ : ConanDependencies = loader.load_dataclass(ConanDependencies, path, obj_file.read(), local_types)

            if obj_templ is None:
                return False

            context.conan.yml_files = obj_templ.targets

        return True

    if len(context.conan.yml_files) == 0:
        if not load_deps():
            return False

    dep_targets = context.conan.yml_files

    if len(dep_targets) != 0:
        context.conan.dependencies = load_dataclass_list(Target, dep_targets, '', context.project_fs, loader, local_types)

    return True


def validate_project(context: CommandContext):
    vctx = ValidationContext()
    vctx.error_format = base.EErrorFormat.MSVC if context.command_args.ide else base.EErrorFormat.Pretty

    validate_targets(vctx, context.project_template.targets + context.conan.dependencies)
    validate_cli_tools(vctx, context.project_template.targets, context.project_template.cli_tools)

    return True


class Generate(Command):
    @property
    def name(self):
        return 'gen'

    @property
    def desc(self):
        return 'Discover toolchains, Resolve dependencies, Generate project files.'

    @property
    def help(self):
        return self.desc

    def ParseArgs(self, arg_parser: ArgumentParser):
        pass

    def OnExecute(self, context: CommandContext):
        # venv could have been generated by conan install or already present in fs
        context.resolve_venv()

        # discover dependencies
        execute_stage(lambda: load_dependencies(context),
                    'Conan dependencies loaded',
                    EErrorCode.UNABLE_TO_LOAD_DEPENDENCIES)

        # Validate target references
        execute_stage(lambda: validate_project(context),
                    'Targets validated',
                    EErrorCode.UNABLE_TO_VALIDATE_TARGETS)

        resolve_context = ResolveContext.from_base(context)
        resolve_context.project=context.project_template

        if len(context.conan.dependencies) > 0:
            # create variants of conan deps per configuration
            context.conan.resolved_targets = execute_stage(lambda: resolve_target_variants(resolve_context, context.config_matrix, context.conan.dependencies),
                                        'Resolved Conan target variants',
                                        EErrorCode.UNABLE_TO_RESOLVE_VARIANTS)

        # create variants of targets per configuration
        context.resolved_targets = execute_stage(lambda: resolve_target_variants(resolve_context, context.config_matrix, context.project_template.targets),
                                    'Resolved target variants dependencies',
                                    EErrorCode.UNABLE_TO_RESOLVE_VARIANTS)

        emit_context = EmitContext.from_base(context)
        emit_context.project_template=context.project_template
        emit_context.resolved_targets=context.resolved_targets
        emit_context.config_matrix=context.config_matrix
        emit_context.conan=context.conan
        emit_context.mem_fs=fs_create('mem://', default_protocol='mem')

        # fill objects for csproj and sln from target variants
        # save to virtual filesystem
        emit_result_tuple = execute_stage(lambda: emit_project(
                                    emit_context,
                                    emitter=FastbuildEmitter()),
                                    'Project data generated',
                                    EErrorCode.UNABLE_TO_EMIT_PROJECT)

        context.fastbuild = Fastbuild(*emit_result_tuple)

        # commit virtual filesystem to os
        if not context.command_args.dry:
            execute_stage(lambda: commit_to_filesystem(context.project_fs, context.fastbuild.emit_fs, context.fastbuild.bff_dir),
                'Files commited to filesystem',
                EErrorCode.UNABLE_TO_COMMIT_FS)

            if check_fastbuild_present(context.fs_venv):
                execute_stage(lambda: call_fbuild_projects(context.fs_root, context.fastbuild.bff_dir, context.fastbuild.bff_file,
                    context.fs_venv),
                    'FASTbuild projects generated',
                    EErrorCode.UNABLE_TO_CREATE_VS_PROJECTS)

                execute_stage(lambda: call_fbuild_solution(context.fs_root, context.fastbuild.bff_dir, context.fastbuild.bff_file,
                    context.project_template.project.project, context.fs_venv),
                    'FASTbuild solution generated',
                    EErrorCode.UNABLE_TO_CREATE_VS_SOLUTION)



from typing import List, Optional, Any, Dict
from dataclasses import dataclass, field, fields, is_dataclass
import os
import fnmatch, typing_inspect
from zetsubou.project.model.config_string import EDefaultConfigSlots
from zetsubou.project.model.configuration import Configuration
from zetsubou.project.model.filter import TargetFilter

from zetsubou.project.base_context import BaseContext
from zetsubou.project.model.kind import ETargetKind
from zetsubou.project.config_matrix import ConfigVariant, ConfigMatrix
from zetsubou.project.runtime.project_loader import ProjectTemplate
from zetsubou.project.model.target import PathList, Target, TargetData, Source, TargetReference, PropertyList
from zetsubou.project.model.toolchain import Toolchain


def PrintDataclass(clazz, id = 0):
    prefix = ''
    for _ in range(id):
        prefix += ' '

    if is_dataclass(clazz):
        print(f'{prefix}Class: {clazz.__name__}')
        for f in fields(clazz):
            print(f'{prefix}name: {f.name}, type: {f.type}')
            PrintDataclass(f.type, id + 4)
    else:
        if typing_inspect.is_generic_type(clazz):
            if typing_inspect.get_origin(clazz) == list:
                subtype = clazz.__args__[0]
                print(f'{prefix}subtype: {subtype}')
                PrintDataclass(subtype, id + 4)
            else:
                print(f'{prefix }Unhandled type {clazz}')


def config_matches_variant(config: str, config_matrix: ConfigVariant) -> bool:
    if config == '*':
        return True

    conf_lower = config.lower()
    matrix_lower = config_matrix.config_string.lower()
    return fnmatch.fnmatch(matrix_lower, conf_lower)


def config_matches_filter(filter: TargetFilter, config_variant : ConfigVariant, target : Target, config: Configuration, toolchain:Toolchain) -> bool:
    if (
            config_matches_variant(filter.config_string, config_variant) and
            (filter.kind is None or target.config.kind in filter.kind) and
            # TODO: FIXME: This is windows only!!!
            (filter.runtime_library is None or toolchain.profile.runtime in filter.runtime_library) and
            (filter.base_configuration is None or config.base_configuration in filter.base_configuration)
       ):
        return True
    return False


def append_list(target: Optional[List[Any]], source: Optional[List[Any]]):
    if target is None:
        raise ValueError('append_list target must not be None!')
    if source is not None:
        target += source


def append_source(target: Source, source: Optional[Source]):
    if source is not None:
        append_list(target.paths, source.paths)
        append_list(target.patterns, source.patterns)


def append_properties(target: PropertyList, source: Optional[PropertyList]):
    if source is not None:
        append_list(target.interface, source.interface)
        append_list(target.public, source.public)
        append_list(target.private, source.private)


@dataclass
class TargetVariant(TargetData):
    config_variant: ConfigVariant = None
    compiler: str = ''
    platform: str = ''

    def __init__(self, config_variant: ConfigVariant):
        TargetData.__init__(self)
        self.config_variant = config_variant

        def init_sources(src: Source):
            src.paths = []
            src.patterns = []

        def init_properties(prop: PropertyList):
            prop.interface = []
            prop.public = []
            prop.private = []

        for data_field in fields(TargetData):
            # Handle Optionals
            field_type = data_field.type
            if typing_inspect.is_optional_type(field_type):
                field_type = field_type.__args__[0]

            self_value = getattr(self, data_field.name)
            self_value = field_type()

            if issubclass(field_type, Source):
                init_sources(self_value)
            else:
                init_properties(self_value)

            setattr(self, data_field.name, self_value)


    def apply(self, data: TargetData):

        for data_field in fields(TargetData):
            # Handle Optionals
            field_type = data_field.type
            if typing_inspect.is_optional_type(field_type):
                field_type = field_type.__args__[0]

            self_value = getattr(self, data_field.name)
            other_value = getattr(data, data_field.name)

            if issubclass(field_type, Source):
                append_source(self_value, other_value)
            elif issubclass(field_type, PathList):
                # TODO: Needs fix, cannot blindly append relative paths
                append_properties(self_value, other_value)
            else:
                append_properties(self_value, other_value)

            setattr(self, data_field.name, self_value)


@dataclass
class ResolvedTarget:
    name: str
    target: Target
    variants: Dict[str, TargetVariant]

    def get_kind(self) -> ETargetKind:
        return self.target.config.kind


@dataclass
class ResolveContext(BaseContext):
    project: ProjectTemplate = None
    target_cache: Dict[str, ResolvedTarget] = field(default_factory=dict)
    resolved_targets: List[ResolvedTarget] = field(default_factory=list)

    @classmethod
    def from_base(cls, source : BaseContext):
        result = ResolveContext()
        result.__dict__.update(source.__dict__)
        return result

    def find_target_variant(self, target: Target, config_str: str) -> Optional[TargetVariant]:
        if target.target in self.target_cache:
            if config_str in self.target_cache[target.target].variants:
                return self.target_cache[target.target].variants[config_str]
        return None

    def add_target_variant(self, target: Target, config_variant: ConfigVariant, target_variant: TargetVariant):
        resolved_target: Optional[ResolvedTarget] = None
        if target.target in self.target_cache:
            resolved_target = self.target_cache[target.target]
        else:
            resolved_target = ResolvedTarget(
                name=target.target,
                target=target,
                variants={}
            )
            self.target_cache[target.target] = resolved_target
            self.resolved_targets.append(resolved_target)
        resolved_target.variants[config_variant.config_string] = target_variant


def append_paths(target: Optional[List[Any]], source: Optional[List[Any]], base_path : str, dep_path : str, fs_root : str):
    if target is None:
        raise ValueError('append_list target must not be None!')
    if source is not None:
        base_fullpath = os.path.normpath(os.path.join(fs_root, base_path))
        dep_fullpath = os.path.normpath(os.path.join(fs_root, dep_path))
        rel_dep_to_base = os.path.relpath(os.path.dirname(dep_fullpath), os.path.dirname(base_fullpath))
        target += [ os.path.join(rel_dep_to_base, i) for i in source ]


def apply_slot(context: ResolveContext, target_variant: TargetVariant, config_variant: ConfigVariant, slot_name: str, slot_value: str, target: Target, config: Configuration, toolchain:Toolchain):
    if EDefaultConfigSlots.toolchain.name == slot_name:
        target_variant.compiler = slot_value

    elif EDefaultConfigSlots.platform.name == slot_name:
        platform = context.project.find_platform(slot_value)
        target_variant.platform = platform.platform
        target_variant.apply(platform)

        for plat_filter in platform.filters:
            if config_matches_filter(plat_filter.filter, config_variant, target, config, toolchain):
                target_variant.apply(plat_filter)

    elif EDefaultConfigSlots.configuration.name == slot_name:
        config = context.project.find_config(slot_value)
        target_variant.apply(config)

        for conf_filter in config.filters:
            if config_matches_filter(conf_filter.filter, config_variant, target, config, toolchain):
                target_variant.apply(conf_filter)


def apply_dependencies_refl(context: ResolveContext, target : Target, target_variant: TargetVariant, config_variant: ConfigVariant, config: Configuration, toolchain:Toolchain):
    target_cls_fields = fields(TargetData)
    target_source_path = target.get_loaded_from_file()

    accessors = [ 'private', 'public', 'interface' ]
    for access in accessors:
        deps : Optional[List[TargetReference]] = getattr(target_variant.dependencies, access)

        if deps is not None:
            for dep_ref in deps:
                lnk_libs = getattr(target_variant.link_libraries, access)
                lnk_libs.append(dep_ref)
                setattr(target_variant.link_libraries, access, lnk_libs)

                dep_variant = create_target_variant(context, dep_ref.target, config_variant, config, toolchain)
                dep_source_path = dep_ref.target.get_loaded_from_file()

                imported_dep_target : bool = dep_ref.target.config.kind == ETargetKind.IMPORTED_TARGET

                if imported_dep_target:
                    # Manually pass imported includes as system ones
                    sys_inc_value_access = getattr(target_variant.system_includes, access)

                    append_paths(sys_inc_value_access, dep_variant.includes.public, target_source_path, dep_source_path, context.fs_root)
                    append_paths(sys_inc_value_access, dep_variant.includes.interface, target_source_path, dep_source_path, context.fs_root)

                    append_paths(sys_inc_value_access, dep_variant.system_includes.public, target_source_path, dep_source_path, context.fs_root)
                    append_paths(sys_inc_value_access, dep_variant.system_includes.interface, target_source_path, dep_source_path, context.fs_root)

                    setattr(target_variant.system_includes, access, sys_inc_value_access)

                for data_field in target_cls_fields:
                    dep_field_holder = getattr(dep_variant, data_field.name)

                    if dep_field_holder is None:
                        continue

                    field_value = getattr(target_variant, data_field.name)

                    # Handle Optionals
                    field_type = data_field.type
                    if typing_inspect.is_optional_type(field_type):
                        field_type = field_type.__args__[0]

                    if issubclass(field_type, PropertyList):
                        dep_field_public = getattr(dep_field_holder, 'public')
                        dep_field_iface = getattr(dep_field_holder, 'interface')
                        field_value_access = getattr(field_value, access)

                        if issubclass(field_type, PathList):
                            # Includes for imported targets have to be handled manually
                            if not (imported_dep_target and data_field.name.find('includes') != -1):
                                append_paths(field_value_access, dep_field_public, target_source_path, dep_source_path, context.fs_root)
                                append_paths(field_value_access, dep_field_iface, target_source_path, dep_source_path, context.fs_root)
                        else:
                            append_list(field_value_access, dep_field_public)
                            append_list(field_value_access, dep_field_iface)

                        setattr(field_value, access, field_value_access)

                    setattr(target_variant, data_field.name, field_value)


def create_target_variant(context: ResolveContext, target: Target, config_variant: ConfigVariant, config: Configuration, toolchain:Toolchain) -> TargetVariant:
    target_variant = context.find_target_variant(target, config_variant.config_string)
    if target_variant is None:
        target_variant = TargetVariant(config_variant)
        target_variant.apply(target)

        for filter_data in target.filters:
            if config_matches_filter(filter_data.filter, config_variant, target, config, toolchain):
                target_variant.apply(filter_data)

        for slot_name, slot_value in config_variant.slots.items():
            apply_slot(context, target_variant, config_variant, slot_name, slot_value, target, config, toolchain)

        for rule in context.project.rules:
            for filter_data in rule.filters:
                if config_matches_filter(filter_data.filter, config_variant, target, config, toolchain):
                    target_variant.apply(filter_data)

        apply_dependencies_refl(context, target, target_variant, config_variant, config, toolchain)

    context.add_target_variant(target, config_variant, target_variant)
    return target_variant


def resolve_target_variants(context: ResolveContext, config_matrix: ConfigMatrix, targets: List[Target]):
    for target in targets:
        for config_variant in config_matrix.variants:
            toolchain = context.project.find_toolchain(config_variant.get_slot(EDefaultConfigSlots.platform), config_variant.get_slot(EDefaultConfigSlots.toolchain))
            config = context.project.find_config(config_variant.get_slot(EDefaultConfigSlots.configuration))
            create_target_variant(context, target, config_variant, config, toolchain)

    return context.resolved_targets



from typing import Optional
from zetsubou.project.model.platform_enums import EArch
from zetsubou.project.model.runtime_library import ERuntimeLibrary
from zetsubou.utils.error import ProjectError

def arch_to_msvc_platform(arch : EArch):
    lookup = {
        EArch.x64 : 'x64',
        EArch.x86 : 'Win32'
    }

    if arch not in lookup:
        raise ProjectError(f"Arch '{arch}' is not supported for MSVC!")

    return lookup.get(arch)

# def runtime_from_msvc_legacy(runtime:str) -> Optional[ERuntimeLibrary]:
#     return {
#         'MD' : ERuntimeLibrary.DYNAMIC_RELEASE,
#         'MDd' : ERuntimeLibrary.DYNAMIC_DEBUG,
#         'MT' : ERuntimeLibrary.STATIC_RELEASE,
#         'MTd' : ERuntimeLibrary.STATIC_DEBUG
#     }.get(runtime, None)



from enum import Enum, auto


class ECompilerFamily(Enum):
    MSVC     = auto()
    CLANG    = auto()
    CLANG_CL = auto()
    GCC      = auto()
    CUSTOM   = auto()


class ELibrarianFamily(Enum):
    MSVC = auto()
    LLVM = auto()
    GCC  = auto()


class ELinkerFamily(Enum):
    LINK     = auto() # MSVC
    LLD_LINK = auto()
    LD_LLD   = auto()
    LD       = auto() # GCC


class ECppStandard(Enum):
    cpp11 = auto()
    cpp14 = auto()
    cpp17 = auto()
    cpp20 = auto()
    cpp23 = auto()
    latest = auto()

    @staticmethod
    def default():
        return ECppStandard.cpp20



import os

from zetsubou.utils.subprocess import call_process

def find_vswhere():
    executable_name = 'vswhere.exe'
    default_installator_path = 'Microsoft Visual Studio/Installer'
    paths = [
        os.path.join(os.environ['ProgramW6432'], default_installator_path),
        os.path.join(os.environ['ProgramFiles(x86)'], default_installator_path)
    ]

    for path in paths:
        local_path = os.path.join(path, executable_name)
        if os.path.exists(local_path):
            return local_path

    return None

def call_vswhere(executable: str):
    args = [
        executable,
        '-prerelease',
        '-format', 'json', '-utf8'
    ]
    out, err = call_process(args, capture=True, realtime=False)
    if err is not None:
        return None
    return out



import os
from dataclasses import dataclass
from typing import List, Optional, Dict

import fs
from fs.base import FS

import importlib_resources as resources
import bentoudev.dataclass.base as base

from zetsubou.project.model.cli_tool import CommandlineTool
from zetsubou.project.model.config_string import EDefaultConfigSlots
from zetsubou.project.model.configuration import Configuration
from zetsubou.project.model.kind import ETargetKind
from zetsubou.project.model.platform import Platform
from zetsubou.project.model.project_template import Project
from zetsubou.project.model.toolchain_profile import Profile
from zetsubou.project.model.rule import Rule
from zetsubou.project.model.target import Target, find_target
from zetsubou.project.model.toolchain import ToolchainDefinition, Toolchain
from zetsubou.utils import logger, yaml_loader
from zetsubou.utils.common import null_or_empty
from zetsubou.utils.error import ProjectError


class NoTraceError(Exception):
    pass


@dataclass
class ExternalConfig:
    profile_file: Optional[str] = None


@dataclass
class ProjectTemplate:
    project: Project
    profile: Optional[Profile]
    platforms: List[Platform]
    system_toolchains: List[ToolchainDefinition]
    platform_toolchains: Dict[str, List[Toolchain]]
    cli_tools: List[CommandlineTool]
    configurations: List[Configuration]
    rules: List[Rule]
    targets: List[Target]

    def find_target(self, name:str):
        return find_target(self.targets, name)

    def find_platform(self, name : str):
        for plat in self.platforms:
            if plat.platform == name:
                return plat
        return None

    def find_config(self, name : str):
        for conf in self.configurations:
            if conf.configuration == name:
                return conf
        return None

    def find_cli_tool(self, name : str):
        for tool in self.cli_tools:
            if tool.name == name:
                return tool
        return None

    def find_toolchain(self, platform_name: str, toolchain_name: str):
        toolchain_list : List[Toolchain] = self.platform_toolchains.get(platform_name, None)
        if toolchain_list is None:
            return None

        for tool in toolchain_list:
            if toolchain_name == tool.name:
                return tool
        return None

    def list_configurations(self):
        result = []
        for conf in self.configurations:
            result.append(conf.configuration)
        return result

    def list_platforms(self):
        result = []
        for plat in self.platforms:
            result.append(plat.platform)
        return result

    # This outputs invalid values - toolchains are dependent on their platforms!
    def list_toolchain(self):
        result = []
        for _, tools in self.platform_toolchains.items():
            names = [ t.name for t in tools ]
            result.extend(names)
        return result

    def slot_values(self):
        result = {
            EDefaultConfigSlots.platform.name : self.list_platforms(),
            EDefaultConfigSlots.configuration.name : self.list_configurations(),
            EDefaultConfigSlots.toolchain.name : self.list_toolchain()
        }

        for option in self.project.options:
            result[option.name] = option.values

        return result

    @staticmethod
    def compile_target_variant_name(kind: ETargetKind, name: str, config_string: str = ''):
        parts = []

        if kind != ETargetKind.INVALID:
            parts.append(f'{name}_{kind.name}')
        else:
            parts.append(name)

        if not null_or_empty(config_string):
            parts.append(config_string)

        return '-'.join(parts)

    @staticmethod
    def decompose_variant_name(variant:str, template_str:str):
        parts = variant.split('-')
        target = parts[0]
        config_parts = parts[1:]

        slots = template_str.split('-')

        if len(config_parts) != len(slots):
            return None, None

        result = {}
        for slot, val in zip(slots, config_parts):
            slot_name = slot.strip('}{')
            result[slot_name] = val

        return target, result





# def load_dataclass_single(clazz: type, obj_ref: Optional[str], proj_dir : str, project_fs : FS, loader, local_types):
#     if obj_ref is None:
#         return None

#     error_format = base.EErrorFormat.MSVC if logger.IsIde() else base.EErrorFormat.Pretty

#     if not base.is_loaded_from_file(clazz):
#         clazz = base.loaded_from_file(clazz)

#     obj_path = fs.path.join(proj_dir, obj_ref)

#     if not project_fs.exists(obj_path) or not project_fs.isfile(obj_path):
#         raise ProjectError(f"Unknown to locate file '{obj_path}'")

#     with project_fs.open(obj_path, 'r', encoding='utf-8') as obj_file:
#         obj_templ = loader.load_dataclass(clazz, obj_path, obj_file.read(), local_types, error_format=error_format)

#         if obj_templ is not None:
#             if not base.is_loaded_from_file(obj_templ):
#                 raise ProjectError(f"Unable to load class '{type(obj_templ)}' from file, missing attribute 'loadable_from_file'!")

#             if logger.IsVisible(logger.ELogLevel.Verbose):
#                 logger.Success(f"Loaded [{clazz.__name__}] '{obj_path}'")

#             obj_templ.set_loaded_from_file(obj_path)
#             return obj_templ
#         else:
#             raise NoTraceError()


def load_dataclass_list(clazz: type, obj_ref_list: Optional[List[str]], proj_dir : str, project_fs : FS, loader, local_types):
    result = []

    error_format = base.EErrorFormat.MSVC if logger.IsIde() else base.EErrorFormat.Pretty

    if not base.is_loaded_from_file(clazz):
        clazz = base.loaded_from_file(clazz)

    if obj_ref_list is not None and len(obj_ref_list) > 0:

        for obj_ref in obj_ref_list:
            obj_path = fs.path.join(proj_dir, obj_ref)

            if not project_fs.exists(obj_path) or not project_fs.isfile(obj_path):
                raise ProjectError(f"Unknown to locate file '{obj_path}'")

            with project_fs.open(obj_path, 'r', encoding='utf-8') as obj_file:
                obj_templ = loader.load_dataclass(clazz, obj_path, obj_file.read(), local_types, error_format=error_format)

                if obj_templ is not None:
                    if not base.is_loaded_from_file(obj_templ):
                        raise ProjectError(f"Unable to load class '{type(obj_templ)}' from file, missing attribute 'loadable_from_file'!")

                    if logger.IsVisible(logger.ELogLevel.Verbose):
                        logger.Success(f"Loaded [{clazz.__name__}] '{obj_path}'")

                    obj_templ.set_loaded_from_file(obj_path)
                    result.append(obj_templ)
                else:
                    raise NoTraceError()

    return result


def load_bundled_rules(loader, local_types) -> List[Rule]:
    result = []
    for res_path in [ 'MsvcRules.yml', 'ClangRules.yml' ]:
        res_content = resources.files('zetsubou.data.rules').joinpath(res_path).read_text()
        obj_templ = loader.load_dataclass(Rule, res_path, res_content, local_types)

        if obj_templ is not None:
            if not base.is_loaded_from_file(obj_templ):
                raise ProjectError(f"Unable to load class '{type(obj_templ)}' from file, missing attribute 'loadable_from_file'!")

            if logger.IsVisible(logger.ELogLevel.Verbose):
                logger.Success(f"Loaded [Rule] '{res_path}'")

            obj_templ.set_loaded_from_file(res_path)
            result.append(obj_templ)
    return result


def load_project_template(project_fs: FS, fs_root : str, filename: str, proj_file_content: str, *, loader=yaml_loader.YamlDataclassLoader(), external_config:Optional[ExternalConfig]=None) -> Optional[ProjectTemplate]:
    try:
        local_types = base.get_types_from_modules([__name__, 'zetsubou.project.model.target'])

        proj_dir = os.path.dirname(filename)
        proj_templ: Project = loader.load_dataclass(Project, filename, proj_file_content, local_types)
        if proj_templ is None:
            return

        proj_templ.set_loaded_from_file(filename)

        def load_dataclasses(clazz: type, obj_ref_list: Optional[List[str]]):
            return load_dataclass_list(clazz, obj_ref_list, proj_dir, project_fs, loader, local_types)

        targets = load_dataclasses(Target, proj_templ.targets)

        proj = ProjectTemplate(
            project=proj_templ,
            profile=None,
            platforms=load_dataclasses(Platform, proj_templ.config.platforms),
            system_toolchains=[],
            platform_toolchains={},
            cli_tools=load_dataclasses(CommandlineTool, proj_templ.config.cli_tools),
            configurations=load_dataclasses(Configuration, proj_templ.config.configurations),
            rules=load_dataclasses(Rule, proj_templ.config.rules),
            targets=targets,
        )

        # Profile handling
        if external_config is not None and external_config.profile_file is not None:
            proj.profile = load_dataclasses(Profile, [ external_config.profile_file ])[0]
        else:
            proj.profile = load_dataclasses(Profile, [ proj.project.config.dev_profile ])[0]

        # Filter by config
        if proj.profile is not None and proj.profile.build_type is not None:
            def is_from_base(config:Configuration):
                return config.base_configuration == proj.profile.build_type

            proj.configurations = list(filter(is_from_base, proj.configurations))
            if len(proj.configurations) == 0:
                raise ProjectError(f"No configurations of base type '{proj.profile.build_type}' found!")
            else:
                proj.configurations = [ proj.configurations[0] ]

        proj.rules += load_bundled_rules(loader, local_types)

        return proj

    except NoTraceError:
        logger.CriticalError(f'Failed to load project \'{filename}\'')
        return None

    except base.DataclassLoadError as error:
        logger.Error(error)
        logger.CriticalError(f'Failed to load project \'{filename}\'')
        return None

    except ProjectError as error:
        logger.Error(error)
        logger.CriticalError(f'Failed to load project \'{filename}\'')
        return None

    except Exception as error:
        logger.Exception(error)
        logger.CriticalError(f'Failed to load project  \'{filename}\'')
        return None


def load_project_from_file(project_fs: FS, fs_root : str, filename: str, *, external_config:Optional[ExternalConfig]=None) -> Optional[ProjectTemplate]:
    if not project_fs.exists(filename):
        logger.CriticalError(f"Unable to locate file '{filename}'")
        return None
    with project_fs.open(filename, 'r', encoding='utf-8') as proj_file:
        return load_project_template(project_fs, fs_root, filename, proj_file.read(), external_config=external_config)



from typing import List
from dataclasses import dataclass, field
from zetsubou.project.model.filter import TargetFilter
from zetsubou.project.model.target import TargetData
from bentoudev.dataclass.base import loaded_from_file


@dataclass
class RuleData(TargetData):
    filter: TargetFilter = field(default_factory=TargetFilter)


@dataclass
@loaded_from_file
class Rule:
    rule: str
    filters: List[RuleData]



from typing import List, Optional
from dataclasses import dataclass, field

from bentoudev.dataclass.base import loaded_from_file, track_source


@dataclass
@track_source
class ProjectConfig:
    verbose_build: Optional[bool] = False
    dev_profile: str = ''
    platforms: List[str] = field(default_factory=list)
    configurations: List[str] = field(default_factory=list)
    cli_tools: Optional[List[str]] = field(default_factory=list)
    # slots: Optional[List[str]] = field(default_factory=list)
    rules: Optional[List[str]] = field(default_factory=list)
    config_string: Optional[str] = '{platform}-{configuration}-{toolchain}'
    venv: Optional[str] = ''


@dataclass
class Option:
    name: str
    values: List[str]


@dataclass
class Conan:
    build_tools: Optional[str] = None
    dependencies: Optional[str] = None
    build: Optional[str] = 'missing'


@loaded_from_file
@dataclass
class Project:
    project: str
    config: ProjectConfig
    targets: List[str]
    conan: Optional[Conan] = None
    options: Optional[List[Option]] = field(default_factory=list)



import os
from typing import Optional
from conan import ConanFile
from conan.tools.files import save


def emit_list(lst):
    return '\n       - '.join(lst)


def emit_property(out, prop, tmpl):
    if prop is not None and len(prop) > 0:
        out.append(tmpl.format(data = emit_list(prop)))


def fix_no_ext(libs, paths):
    results = []
    for path in paths:
        for _, _, filenames in os.walk(path):
            for file in filenames:
                for lib in libs:
                    if file.find(lib) != -1:
                        results.append(os.path.basename(file))
    return results


def _emit_cpp_info(dep_cpp_info):
    sections = []
    emit_property(sections, dep_cpp_info.includedirs,
        '    includes:\n'
        '       interface:\n'
        '       - {data}\n')

    emit_property(sections, dep_cpp_info.defines,
        '    defines:\n'
        '       interface:\n'
        '       - {data}\n')

    emit_property(sections, dep_cpp_info.cxxflags,
        '    compiler_flags:\n'
        '       interface:\n'
        '       - {data}\n')

    emit_property(sections, fix_no_ext(dep_cpp_info.libs, dep_cpp_info.libdirs),
        '    link_libraries:\n'
        '       interface:\n'
        '       - {data}\n')

    emit_property(sections, dep_cpp_info.libdirs,
        '    linker_paths:\n'
        '       interface:\n'
        '       - {data}\n')

    emit_property(sections, dep_cpp_info.sharedlinkflags + dep_cpp_info.exelinkflags,
        '    linker_flags:\n'
        '       interface:\n'
        '       - {data}\n')

    return sections


def generate_dependencies(conanfile: ConanFile):
    result = {}

    for _, dependency in conanfile.dependencies.items():

        components = []
        for comp_key, component in dependency.cpp_info.components.items():
            if comp_key is None:
                continue

            comp_name = f"{dependency.ref.name}{comp_key}".replace("-", "_")

            components.append(comp_name)

            sections = _emit_cpp_info(component)
            result[f'{comp_name}.part.yml'] = ''.join(sections)

        sections = _emit_cpp_info(dependency.cpp_info)

        emit_property(sections, components,
        '    dependencies:\n'
        '       interface:\n'
        '       - {data}\n')

        dep_name = dependency.ref.name.replace("-", "_")
        result[f'{dep_name}.part.yml'] = ''.join(sections)

    return result


def generate(conanfile:ConanFile, output_folder:Optional[str] = None):
    generator_files = generate_dependencies(conanfile)

    generate_folder = conanfile.folders.generators_folder if output_folder is None else output_folder

    for generator_file, content in generator_files.items():
        save(conanfile, os.path.join(generate_folder, generator_file), content)


# Conan 2.0 - Deployer entrypoint
def deploy(graph, output_folder: str, **kwargs):
    conanfile = graph.root.conanfile
    generate(conanfile, output_folder)


# Used to inject deployer into conan
def get_conan_deployer_path():
    return os.path.realpath(__file__)



from typing import List, Optional
from dataclasses import dataclass


class AdditionalSolutionOptions:
    # (optional) Project(s) set to build when "Build Solution" is selected
    SolutionBuildProject   : Optional[List[str]]
    # (optional) Project(s) set deploy
    SolutionDeployProjects : Optional[List[str]]

@dataclass
class SolutionConfigs(AdditionalSolutionOptions):
    # Platform(s) (default "Win32", "x64")
    Platform : str
    # Config(s) (default "Debug", "Release")
    Config : List[str]
    # (optional) Solution Config
    SolutionConfig : Optional[str]
    # (optional) Solution Platform
    SolutionPlatform : Optional[str]

@dataclass
class SolutionFolders:
    # Folder path in Solution
    Path : str
    # (optional) Project(s) to include in this folder
    Projects : Optional[List[str]]
    # (optional) Solution Item(s) (files) to include in this folder
    Items : Optional[List[str]]

@dataclass
class SolutionDependencies:
    # Project(s) to specify dependencies for
    Projects : List[str]
    # Project(s) the above projects depend on
    Dependencies : List[str]


class VSSolution(AdditionalSolutionOptions):
    # (optional) Alias
    Alias : Optional[str]

    ###############################################
    # Basic options
    ###############################################

    # Path to Solution file to be generated
    SolutionOutput : str
    # (optional) Project(s) to include in Solution
    SolutionProjects : Optional[List[str]]
    # (optional) Solution configurations (see below)
    SolutionConfigs : Optional[List[SolutionConfigs]]

    ###############################################
    # Folders
    ###############################################

    # (optional) Folders to organize projects (see below)
    SolutionFolders : Optional[List[SolutionFolders]]

    ###############################################
    # Advanced options
    ###############################################

    # (optional) Project dependency information (see above)
    SolutionDependencies : Optional[List[SolutionDependencies]]

    ###############################################
    # Version Info  
    ###############################################

    # (optional) Version of Solution (default "14.0.22823.1" VS2015 RC)
    SolutionVisualStudioVersion : Optional[str]
    # (optional) Min version of Solution (default "10.0.40219.1" VS2010 Express)
    SolutionMinimumVisualStudioVersion : Optional[str]

    @staticmethod
    def keyword() -> str:
        return "VSSolution"



from dataclasses import dataclass, field
from typing import List, Optional
from zetsubou.project.model.virtual_environment import VirtualEnvironment
from zetsubou.utils.subprocess import call_process_venv


@dataclass
class CommandlineTool:
    name: str
    executable: str

    """
    Supports:
        %1 - Input file(s) as provided via ExecInput argument.
        %2 - Output file as provided by output_file argument.
        %3 - Output dir as provided by output_dir argument.
    """
    options: List[str]
    output_file: str
    output_dir: str

    _exec_fullpath: Optional[str] = field(init=False, repr=False, default=None)

    # Can only be called after tool fullpath was resolved successfully
    @property
    def executable_fullpath(self):
        assert self._exec_fullpath is not None
        return self._exec_fullpath

    # Finds executable fullpath in virtual environment
    def resolve(self, venv:VirtualEnvironment) -> bool:
        # Downside is that we kinda assumed venv would be a thing, but it is not - fastbuild doesn't want it...
        out, err = call_process_venv(['where', self.executable], venv, capture=True, realtime=False)

        if err is not None:
            return False

        self._exec_fullpath = out.strip()
        return True



import time

from zetsubou.utils import logger
from zetsubou.utils.busy_indicator import BusyIndicator
from zetsubou.utils.error_codes import EErrorCode, ReturnErrorcode

def execute_stage(func, succ_msg: str, err_code: EErrorCode):
    indicator_disabled = logger.IsVisible(logger.ELogLevel.Verbose) or logger.IsIde()

    start_time = time.time()

    with BusyIndicator(not indicator_disabled):
        result = func()

    end_time = time.time()

    if result is not None and result is not False:
        elapsed = end_time - start_time
        resolution = 2 if elapsed > 0.01 else 3
        logger.Success(f'{succ_msg} - {elapsed:.{resolution}f}sec')
        return result
    else:
        raise ReturnErrorcode(err_code)



from typing import List
from sys import platform
import os


class Version:
    major: int = 0
    minor: int = 0
    path: int = 0

    def __init__(self, ver_str:str):
        ver_parts = ver_str.split('.')

        num_sub_ver = len(ver_parts)
        if num_sub_ver == 0:
            return

        if num_sub_ver >= 1:
            self.major = int(ver_parts[0])

        if num_sub_ver >= 2:
            self.minor = int(ver_parts[1])

        if num_sub_ver >= 3:
            self.path = int(ver_parts[2])

    def __str__(self):
        if self.path != 0:
            return f"{self.major}.{self.minor}.{self.path}"
        if self.minor != 0:
            return f"{self.major}.{self.minor}"
        return str(self.major)

    def __eq__(self, other):
        return self.major == other.major and self.minor == other.minor and self.path == other.path

    def __lt__(self, other):
        if self.major < other.major:
            return True
        elif self.major > other.major:
            return False

        if self.minor < other.minor:
            return True
        elif self.minor > other.minor:
            return False

        if self.path < other.path:
            return True
        elif self.path > other.path:
            return False

        return False


def foreach(lst, fn):
    if lst is None:
        return
    if len(lst) == 0:
        return
    for element in lst:
        fn(element)


def filter_none(lst):
    return filter(lambda i : i is not None, lst)


def is_in_enum(value, enum_class):
    if isinstance(value, enum_class):
        return True
    return value in enum_class.__members__


# Find more appropriate way to do this...
def fix_path(path : str):
    path = path.replace('\\\\', '/')
    path = path.replace('\\', '/')
    return path


def fix_path_win32(path : str):
    path = path.replace('/', '\\')
    return path


def fix_path_os(path : str):
    posix_path = fix_path(path)
    if platform == 'win32':
        return fix_path_win32(posix_path)
    return posix_path


def get_subdirs(dir: str):
    if not os.path.exists(dir):
        return []
    return [f.name for f in os.scandir(dir) if f.is_dir]


def get_files(dir: str):
    if not os.path.exists(dir):
        return []
    return [f.name for f in os.scandir(dir) if f.is_file]


def get_env_path() -> List[str]:
    env_path_str = os.getenv('PATH')
    return env_path_str.split(';')


# folder/subfolder/whatever/file.ext1.ext2 -> file
def filename_no_ext(path:str):
    base = os.path.basename(path)
    return base[:base.find('.')]


def null_or_empty(txt : str):
    return txt is None or txt == ''


def split(pred, collection):
    a = []
    b = []
    for element in collection:
        if pred(element):
            a.append(element)
        else:
            b.append(element)
    return a, b


# TODO replace with *args, this inline temporary array is unnecessary
def join_unique(lists : List[List[str]]):
    unique = list()
    for entry in lists:
        unique.extend(j for j in entry if j not in unique)
    return unique


def list_to_string(lists : List[str]):
    return f" {' '.join(lists)}"



import threading
import sys
import time


class BusyIndicator:
    _cancelation_token = False
    _thread = None

    def __init__(self, enabled:bool=True):
        self._enabled = enabled

    def __enter__(self):
        if not self._enabled:
            return

        self._thread = threading.Thread(target=BusyIndicator._worker, args=[lambda: self._cancelation_token])
        self._thread.start()

    @staticmethod
    def _worker(cancelation_token):
        symbols = ['-', '/', '|', '\\', '-']
        idx = 0
        start_time = time.time()

        time.sleep(0.2)

        while not cancelation_token():
            elapsed = time.time() - start_time
            if elapsed > 1.5:
                sys.stdout.write(f"\r   {symbols[idx]} {elapsed:.1f}sec")
            else:
                sys.stdout.write(f"\r   {symbols[idx]}")
            idx += 1
            if idx >= len(symbols):
                idx = 0
            time.sleep(0.1)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._cancelation_token = True



# pyright: reportMissingImports=false
# pylint: disable=import-error, bare-except
from typing import Optional, List, Union
from dataclasses import dataclass, field

# This file can be imported from conan generator
# So it needs to support importing via absolute and relative packages
try:
    from zetsubou.project.model.configuration_enums import EBaseConfiguration
    from zetsubou.project.model.platform_enums import EArch, EVersionSelector
    from zetsubou.project.model.runtime_library import ERuntimeLibrary
    from zetsubou.project.model.toolchain_enums import ECompilerFamily, ECppStandard
except:
    pass


try:
    from configuration_enums import EBaseConfiguration
    from platform_enums import EArch, EVersionSelector
    from runtime_library import ERuntimeLibrary
    from toolchain_enums import ECompilerFamily, ECppStandard
except:
    pass


@dataclass
class IToolchainProfile:
    target_arch: Optional[EArch] = None
    compiler_family: Optional[ECompilerFamily] = None
    compiler_version: Optional[Union[EVersionSelector, str]] = None
    cppstd: Optional[ECppStandard] = None


@dataclass
class Windows_ToolchainProfile(IToolchainProfile):
    toolset: Optional[Union[EVersionSelector, str]] = None
    runtime: Optional[ERuntimeLibrary] = None


@dataclass
class Linux_ToolchainProfile(IToolchainProfile):
    libcxx: Optional[str] = None


@dataclass
class Profile:
    build_type: Optional[EBaseConfiguration] = None
    toolchains: List[Union[IToolchainProfile, Windows_ToolchainProfile, Linux_ToolchainProfile]] = field(default_factory=list)
